/**
 * Java SendMail
 * Copyright (C) 2005  Ondrej Vondra
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package attachments;

import java.io.*;

public class cAttachment {
  private InputStream in;
  private BufferedReader bufin;

  public String contentType, name, transferEncoding;

  private boolean dobase64;
  final int CHARSPERLINE = 76; // defined in rfc1521

  public cAttachment(String name, InputStream content, String contentType)
  {
    in = content;
    this.contentType = contentType;
    this.name = name;

    if (contentType != null)
      dobase64 = !contentType.matches("text/.*");
    else
    {
      System.out.println("Warning: null Content-Type for attachment "+name);
      contentType = "application/unknown";
      dobase64 = true;
    }
    if (dobase64)
      transferEncoding = "base64";
    else {
      bufin = new BufferedReader(new InputStreamReader(in));
      transferEncoding = "quoted-printable";
    }
  }

  private int byte_to_uint(byte b)
  {
    if (b < 0)
      return 256 + (int)b;
    else
      return b;
  }

  private boolean qp_char_needs_encoding(int ch, boolean lastonline)
  {
    if ((ch == 32) || (ch == 9)) // rule #3 in rfc
      return lastonline;
    if ((ch < 33) || (ch == 61) || (ch > 126)) // rule #1 and #2 in rfc
      return true;
    return false;
  }

  private String qp_encode(int ch) throws Exception
  {
    if (ch > 255) // :)
      throw new Exception("Attempting to encode character > 0xff " + ch);
    String r = "=";

    if ((ch >> 4) > 9)
      r += (char)('A' + (ch >> 4) - 10);
    else
      r += (char)('0' + (ch >> 4));

    if ((ch & 0xf) > 9)
      r += (char)('A' + (ch & 0xf) - 10);
    else
      r += (char)('0' + (ch & 0xf));

    return r;
  }

  byte qp_cached_line[] = null;
  private String readLine_qp() throws Exception // quoted-printable encoding
  {
    String encoded = "";
    int i=0, ch;
    boolean lastonline, needsenc, dosoftbreak=false;

    try { // exceptions need servicing in order to ivalidate cache
      if (qp_cached_line == null)
      {
        String l = bufin.readLine();
        if (l==null)
          return null;
        qp_cached_line = l.getBytes(); // no unicode encoding please
      }

      if (qp_cached_line.length == 0)
      {
        qp_cached_line = null;
        return encoded;
      }
      do
      {
        ch = byte_to_uint(qp_cached_line[i]);

        lastonline = (i == qp_cached_line.length-1);
        needsenc = qp_char_needs_encoding(ch, lastonline);

        // rule #5
        /* we must avoid situations, when we have 76 chars enocoded line
         and we have some characters remaining on the input
         in this case we would have no room to add soft line break (equal sign)
         */
        if (encoded.length() == CHARSPERLINE - 1)
        /* encoded line can contain only one more character */
          if (needsenc || (!lastonline))
          /* we will have to add 3 chars now or more in the next loop so */
          /* do soft line break */
            dosoftbreak = true;

        if (needsenc) // we will encode = add 3 characters
        {
          if (encoded.length() + 3 == CHARSPERLINE)
          /* encoded line will be full */
          {
            if (!lastonline)
              dosoftbreak = true;
          }
          else
          {
            if (encoded.length() + 3 > CHARSPERLINE)
            /* encoded line would overflow */
              dosoftbreak = true;
          }
        }

        // all checks are done so finally we will encode
        if (dosoftbreak)
        {
          byte ncache[] = new byte [qp_cached_line.length - i];
          System.arraycopy(qp_cached_line, i, ncache, 0, ncache.length);
          qp_cached_line = ncache;
          encoded += "=";
          return encoded;
        }

        if (needsenc)
          encoded += qp_encode(ch);
        else
          encoded += (char) ch;

        i ++;
      } while (!lastonline);
    }
    catch (Exception e) {
      qp_cached_line = null;
      throw e;
    }
    qp_cached_line = null;
    return encoded;
  }

  private String readLine_base64() throws Exception
  {
    String line = null;
    int buf[] = new int[3], b[];
    int bi, n, size = CHARSPERLINE;

    while (size > 0)
    /* read 3 bytes as integers - this is a trick for encoder */
    /* shifting in base64 would otherwise fuck up the bits in bytes > 127 */
    /* this however is cool, because all byte values produce unsigned integers*/
    {
      for (bi=0; bi<3; bi++)
      {
        n = in.read();
        if (n == -1)
          break;
        buf[bi] = n;
      }
      if (bi == 0)
        break;

      b = new int[bi];
      System.arraycopy(buf, 0, b, 0, bi);

      if (line==null)
        line = new String("");
      line += cBase64Encoder.encode(b);

      size -= 4;
      if (bi < 3)
        break;
    }
    return line;
  }

  public String readLine() throws Exception
  {
  if (dobase64)
    return readLine_base64();
  return readLine_qp();
  }
}